/* confirm-all: read context filter for detecting RDD sites.
   Detects and filters potentially erroneous RDDs due to errors at positions 1-2 of a read.

   Author: Liliana Florea
   Date: 25 May 2011

   Compile with:
     gcc -O3 confirm-pos.c lib.c -o confirm-pos -lm
*/

#include "libc.h"
#include "lib.h"

#define ORI_MASK       0x10

#define DEFAULT_ALLOC    10
#define READ_LEN         50

#define max(x,y)  (((x)>=(y)) ? (x):(y))
#define min(x,y)  (((x)<=(y)) ? (x):(y))

static char Usage[] = "%s sites_file qryfile_sam\n";

#ifdef MOUSE
static char *GAIds[] = {
       "chr1", "chr2", "chr3", "chr4", "chr5", "chr6", "chr7", "chr8", "chr9", "chr10",
       "chr11", "chr12", "chr13", "chr14", "chr15", "chr16", "chr17", "chr18", "chr19", "chrX",
       "chrY", "chrM", "chr1_random", "chr3_random", "chr4_random", "chr5_random", "chr7_random",
       "chr8_random", "chr9_random", "chr13_random", "chr16_random", "chr17_random",
       "chrUn_random", "chrX_random", "chrY_random"};
static int numGAIds = 35;
#else
static char *GAIds[] = {
       "chr1", "chr2", "chr3", "chr4", "chr5", "chr6", "chr7", "chr8", "chr9", "chr10",
       "chr11", "chr12", "chr13", "chr14", "chr15", "chr16", "chr17", "chr18", "chr19", "chr20", 
       "chr21", "chr22", "chrX", "chrY", "chrM", "chr1_gl000191_random", "chr1_gl000192_random",
       "chr4_gl000193_random", "chr4_gl000194_random", "chr7_gl000195_random",
       "chr8_gl000196_random", "chr8_gl000197_random", "chr9_gl000198_random",
       "chr9_gl000199_random", "chr9_gl000200_random", "chr9_gl000201_random",
       "chr11_gl000202_random", "chr17_gl000203_random", "chr17_gl000204_random",
       "chr17_gl000205_random", "chr17_gl000206_random", "chr18_gl000207_random",
       "chr19_gl000208_random", "chr19_gl000209_random", "chr21_gl000210_random",
       "chrUn_gl000211", "chrUn_gl000212", "chrUn_gl000213", "chrUn_gl000214",
       "chrUn_gl000215", "chrUn_gl000216", "chrUn_gl000217", "chrUn_gl000218",
       "chrUn_gl000219", "chrUn_gl000220", "chrUn_gl000221", "chrUn_gl000222",
       "chrUn_gl000223", "chrUn_gl000224", "chrUn_gl000225", "chrUn_gl000226",
       "chrUn_gl000227", "chrUn_gl000228", "chrUn_gl000229", "chrUn_gl000230",
       "chrUn_gl000231", "chrUn_gl000232", "chrUn_gl000233", "chrUn_gl000234",
       "chrUn_gl000235", "chrUn_gl000236", "chrUn_gl000237", "chrUn_gl000238",
       "chrUn_gl000239", "chrUn_gl000240", "chrUn_gl000241", "chrUn_gl000242",
       "chrUn_gl000243", "chrUn_gl000244", "chrUn_gl000245", "chrUn_gl000246",
       "chrUn_gl000247", "chrUn_gl000248", "chrUn_gl000249"};
static int numGAIds = 84; 
#endif
                       
typedef struct site {
   char   chrom[50];
   int    gaid;
   int    pos;
   char   refbase;
   char   mutbase;
   int    numreads;
   int    refcount, mutcount;
   int    refflag, mutflag;
   char  *text;
} site_t;

static site_t *Sites = NULL;
static int     numSites = 0, maxSites = 0;
static site_t *Caps[10000];
static int     countsCaps[10000]; 


static void processCov(int gaid, int ori, int from, char *cigar, char *aln, site_t **caps, int *capscounts, site_t *sites, int numsites);
static void load_sites(char *);
static int  overlap(int pos, int start, int s1, int g1, int s2, int g2, int s3, int g3, int s4);
static int  overlapX(int pos, int start, char *cigar, char *aln);
static int  lookup_gaid(char *,char **,int);
static int  cmp_sites(const void *,const void *);
static int  lookup_left(int pos, site_t *sites, int l, int u);
static int  lookup_right(int pos, site_t *sites, int l, int u);


static int cmp_sites(const void *A, const void *B)
{
  int tmp;

  if ((tmp=(((const site_t *)A)->gaid - ((const site_t *)B)->gaid)) != 0)
      return tmp;
  else
      return (((const site_t *)A)->pos - ((const site_t *)B)->pos);
}


int main(int argc, char *argv[])
{
   char   buf[10000], *tok, aln[1000], cigar[1000], last_ga[100];
   int    i, code, ori, gaid, from, pos;
   FILE  *fp;
 
   if (argc!=3) fatalf(Usage, argv[0]);

   *last_ga = '\0'; gaid = -1;

   /* make sure the database is sorted by gaid, then increasingly by pos!!! */
   load_sites(argv[1]);
   qsort(Sites, numSites, sizeof(site_t), cmp_sites);
   memset(Caps,0,10000*sizeof(site_t *));
   memset(countsCaps,0,10000*sizeof(int));
   for (i=0; i<numSites; i++) {
     if (Caps[Sites[i].gaid] == NULL)
        Caps[Sites[i].gaid] = Sites+i;
     countsCaps[Sites[i].gaid]++;
   }

   fprintf(stderr, "Loaded sites...\n");

   /* now process SAM alignments, sorted by chromosome; print stats after each chrom batch */
   fp = ckopen(argv[2], "r");

   fprintf(stderr, "#chrom pos refbase mutbase numreads refreads refflags mutreads mutflags\n");

   while (fgets(buf,10000,fp)) {
     /* read one SAM alignment:
         HWI-EASXXX_37-2-Arito:7:21:1660:399     0       chr1    3000654 255     36M     *       0       0       ACTGGAGGTTAGCAATCTAGGGTCAGATAATTATAG  hhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh    NM:i:0
     */
     if (buf[0] == '@') continue;
 
     tok = strtok(buf, " \t\n");   /* skip over id */
     tok = strtok(NULL, " \t\n");   
     code = atoi(tok);
     ori = (((code & ORI_MASK) >> 4) ? -1 : 1);

     tok = strtok(NULL, " \t\n");  
     if (!*last_ga || strcmp(last_ga, tok)) {
        if (*last_ga) {
          if (gaid>=0 && Caps[gaid]) {
            for (i=Caps[gaid]-Sites; i<Caps[gaid]+countsCaps[gaid]-Sites; i++) {
              site_t *sptr = Sites+i;

              if (sptr->mutflag!=sptr->mutcount) printf("%s", sptr->text);

              fprintf(stderr, "%s %d %c %c %d %d %d %d %d\n",
                      GAIds[sptr->gaid], sptr->pos, sptr->refbase, sptr->mutbase, sptr->numreads,
                      sptr->refcount, sptr->refflag, sptr->mutcount, sptr->mutflag);

              free(sptr->text); sptr->text = NULL;
            }
            fflush(NULL);
          }
        }
        if ((gaid=lookup_gaid(tok,GAIds,numGAIds))<0) {
           fprintf(stderr, "Warning: gaId %s not found. Skipping aligment.\n", tok); 
           continue;
        }
        strcpy(last_ga, tok);
     }
     tok = strtok(NULL, " \t\n"); 
     from = atoi(tok);
     tok = strtok(NULL, " \t\n");  /* skip over quality score */
     tok = strtok(NULL, " \t\n");  /* read CIGAR alignment */ 
     strcpy(cigar, tok);
     tok = strtok(NULL, " \t\n");  /* skip over col(1) */
     tok = strtok(NULL, " \t\n");  /* skip over col(2) */
     tok = strtok(NULL, " \t\n");  /* skip over col(3) */
     tok = strtok(NULL, " \t\n");  /* read SEQ */
     strcpy(aln, tok);

     processCov(gaid,ori,from,cigar,aln,Caps,countsCaps,Sites,numSites);
   }
   fclose(fp);

   if (gaid>=0 && Caps[gaid]) {
     for (i=Caps[gaid]-Sites; i<numSites; i++) {
       site_t *sptr = Sites+i;

       if (sptr->mutflag != sptr->mutcount) printf("%s", sptr->text);

       fprintf(stderr, "%s %d %c %c %d %d %d %d %d\n", 
               GAIds[sptr->gaid], sptr->pos, sptr->refbase, sptr->mutbase, sptr->numreads,
               sptr->refcount, sptr->refflag, sptr->mutcount, sptr->mutflag);

       free(sptr->text); sptr->text = NULL; 
     }
     fflush(NULL);
   }

   if (Sites) free(Sites);

   exit(0);
}


static void load_sites(char *dbfile)
{
   FILE   *fp;
   char    buf[100000], *tok, *s;
   site_t *gr;
   int     i;

   /* chr1    14574   A       21      ggggggggg,ggggg,,,,,,   HHGHHHFHH=IHHHH#HHGGF */

   fp = ckopen(dbfile, "r"); 
   while (fgets(buf, 100000, fp)) {
      if (numSites>=maxSites) {
         maxSites = max(2*maxSites,DEFAULT_ALLOC);
         Sites = realloc(Sites,maxSites*sizeof(site_t));
      }
      gr = Sites+numSites;

      int l = strlen(buf);
      gr->text = malloc(l+2);
      strcpy(gr->text, buf);
      gr->text[l] = '\0';

      tok = strtok(buf, " \n\t");
      strcpy(gr->chrom,tok);
      gr->gaid = lookup_gaid(tok,GAIds,numGAIds);
      if (gr->gaid<0) {
        fprintf(stderr, "Warning: gaId %s not found. Skipping.\n", tok);
        free(gr->text);
        continue;
      }

      tok = strtok(NULL, " \n\t");
      gr->pos = atoi(tok);

      tok = strtok(NULL, " \n\t");
      gr->refbase = *tok;

      tok = strtok(NULL, " \n\t");
      gr->numreads = atoi(tok);

      tok = strtok(NULL, " \n\t");
      for (s=tok; *s; s++) {
        char ch = toupper(*s);
        if ((ch=='A') || (ch=='C') || (ch=='G') || (ch=='T')) {
           gr->mutbase = ch;
           break;
        }
      }

      gr->refcount = gr->mutcount = gr->refflag = gr->mutflag = 0;

      numSites++;
   }
   fclose(fp);

   return;
}

static int lookup_gaid(char *query_gaid, char *gaids[], int numgaids)
{
   int    i;

   for (i=0; i<numgaids; i++)
       if (!strcmp(query_gaid,gaids[i])) return i;

   return -1;
}

static void processCov(int gaid, int ori, int from, char *cigar, char *aln, site_t **caps, int *capscounts, site_t *sites, int numsites)
{
  int i, l, u, to, L, offset;
  int t1, t2, t3, t4, t5, t6, t7;
  char *s, key[10], ki;

  if (caps[gaid]==NULL) return;  /* no sites on that gaid */

#if 1
  *key = '\0';
  to = from; 
  for (s=cigar, ki=0; *s; s++) {
    if (isdigit(*s)) key[ki++] = *s;
    else {
      key[ki] = '\0'; l = atoi(key);
      switch (*s) {
         case 'M': to += l; break;
         case 'N': to += l; break; 
         case 'I': break;
         case 'D': to += l; break;
         default : fprintf(stderr, "Warning: Unrecognized cigar code (%c).\n", *s); break;
      }
      *key = 0; ki = 0; 
    }
  }
  to -= 1;
  L = strlen(aln);
#else
  /* old format, gap-free alignments only */
  t1 = t2 = t3 = t4 = t5 = t6 = t7 = 0;

  if (sscanf(cigar, "%dM%dN%dM%dN%dM%dN%dM", &t1,&t2,&t3,&t4,&t5,&t6,&t7)==5) {
    /* 8M299N18M171N11M80N13M */
    to = from + t1 + t2 + t3 +t4 + t5 + t6 + t7 -1; 
  } else if (sscanf(cigar, "%dM%dN%dM%dN%dM", &t1,&t2,&t3,&t4,&t5)==5) {
    /* 13M1618N48M1995N14M */
    to =  from + t1 + t2 + t3 +t4 + t5 -1;
  } else if (sscanf(cigar, "%dM%dN%dM", &t1, &t2, &t3)==3) {
    to =  from + t1 + t2 + t3 -1;
  } else if (sscanf(cigar, "%dM", &t1)==1) {
    to = from + t1 -1;
  } else {
    fprintf(stderr, "Warning: Unrecognized CIGAR format %s. Skipping.\n", cigar);
  }
  assert(t1+t3+t5+t7==strlen(aln));
  L = strlen(aln);
#endif

  /* only searching for the first or last two letters, so just check the two ends */

  l = lookup_left(from,sites,caps[gaid]-sites,caps[gaid]-sites+capscounts[gaid]-1);
  u = lookup_right(to,sites,caps[gaid]-sites,caps[gaid]-sites+capscounts[gaid]-1); 
  
//printf("L: %d (%s,%d) u: %d (%s,%d)\n", l, sites[l].chrom, sites[l].pos, u, sites[u].chrom, sites[u].pos);

  if ((l<0) || (u<0)) return;

  for (i=l; i<=u; i++) {
    /* count reads */ 
#if 1
    if ((offset=overlapX(sites[i].pos,from,cigar,aln))>=0) {
#else
    /* old, gap-free alignments only */
    if ((offset=overlap(sites[i].pos,from,t1,t2,t3,t4,t5,t6,t7))>=0) {
#endif
#ifdef DEBUG
      printf("OVERLAP: site: %d offset: %d siteref: %c sitemut: %c aln@offset: %c\n", 
              sites[i].pos, offset, sites[i].refbase, sites[i].mutbase, aln[offset]);
#endif
      if (aln[offset]==sites[i].refbase) sites[i].refcount++;
      else if (aln[offset]==sites[i].mutbase) sites[i].mutcount++;
      else fprintf(stderr, "Warning: letter (%c) not reference (%c) or variant (%c). (%s,%d)\n",
                          aln[offset], sites[i].refbase, sites[i].mutbase, sites[i].chrom, sites[i].pos);

#if DEPRECATED
      if ((ori==1 && (!offset || (offset==1))) ||
          (ori==-1 && ((offset==L-1) || (offset==L-2)))) {
#else
      if (!offset || (offset==1) || (offset==L-1) || (offset==L-2)) {
#endif
        if (aln[offset]==sites[i].refbase) sites[i].refflag++;
        else if (aln[offset]==sites[i].mutbase) sites[i].mutflag++;
      }
    } else {
#ifdef DEBUG
      printf("OVERLAP: site: %d offset: %d siteref: %c sitemut: %c aln@offset: na\n",
              sites[i].pos, offset, sites[i].refbase, sites[i].mutbase);
#endif
    }
  }
}

/* NOTE: assume no complications from I/D operations */
static int overlapX(int pos, int start, char *cigar, char *aln) 
{
  int  l, cpos, ki;
  char key[10], *s, *q;

  *key = '\0';
  for (s=cigar, ki=0, cpos=start, q=aln; *s; s++) {
    if (isdigit(*s)) key[ki++] = *s;
    else {
      key[ki] = '\0'; l = atoi(key);
      switch (*s) {
         case 'M': if ((cpos<=pos) && (pos<=cpos+l-1))
                     return ((q+(pos-cpos))-aln);
                   cpos += l; q += l;
                   break;
         case 'N': cpos += l;
                   break;
         case 'I': q += l;
                   break;
         case 'D': cpos += l;
                   break;
         default : fprintf(stderr, "Warning: Unrecognized cigar code (%c).\n", *s);
                   break;
      }
      *key = 0; ki = 0;
    }
  }

  return -1;
}

static int overlap(int pos, int start, int s1, int g1, int s2, int g2, int s3, int g3, int s4)
{
  if ((start<=pos) && (pos<=start+s1-1))
    return pos-start;
  if (s2 && (start+s1+g1<=pos) && (pos<=start+s1+g1+s2-1))
    return pos-start-g1;
  if (s3 && (start+s1+g1+s2+g2<=pos) && (pos<=start+s1+g1+s2+g2+s3-1))
    return pos-start-g1-g2;
  if (s4 && (start+s1+g1+s2+g2+s3+g3<=pos) && (pos<=start+s1+g1+s2+g2+s3+g3+s4-1))
    return pos-start-g1-g2-g3;

  return -1;
}
 
/* find the first one that is larger or equal */
static int lookup_left(int pos, site_t *sites, int l, int u)
{
  int m;

  if (u-l<=1) {
    if (sites[l].pos>=pos) return l;
    else if (sites[u].pos>=pos) return u;
    else return -1;
  }

  m = (int)((l+u)/2);
  if (pos==sites[m].pos) return m;
 
  if (sites[m].pos<pos) return lookup_left(pos,sites,m+1,u);
  else lookup_left(pos,sites,l,m);
}

/* find the last one that is smaller or equal */
static int lookup_right(int pos, site_t *sites, int l, int u)
{
  int m;

  if (u-l<=1)  {
    if (sites[u].pos<=pos) return u;
    else if (sites[l].pos<=pos) return l;
    else return -1;
  }

  m = (int)((l+u)/2);
  if (pos==sites[m].pos) return m;
  else if (pos<sites[m].pos) lookup_right(pos,sites,l,m-1);
  else lookup_right(pos,sites,m,u);
}
