#!/usr/bin/perl;
use strict;

$| = 1;

($#ARGV>=3) or die
"Usage: $0 bamfile genome bwtindex syscall \n" .
"          [-annot annotfile] [-snp snpfile] [-dir workdir] [-verbose]\n" .
"NOTE: Give full paths for bamfile, genome, bwtindex, syscall.\n";


my $R = 20;
my $P = 20;

my $BAMFILE  = shift;
my $GENOME   = shift;
my $BWTINDEX = shift;
my $SYSCALL  = shift;
my $verbose  = 0;
my $work = "./";

(($BAMFILE=~/^\//) && ($GENOME=~/^\//) && ($BWTINDEX=~/^\//) && ($SYSCALL=~/^\//))
   or die "Error: Full paths needed for bamfile, genome, bowtie_index and syscall directory.\n";
(-e $BAMFILE) or die "$BAMFILE: bamfile does not exist.\n";
(-e $GENOME) or die "$GENOME: genome file does not exist.\n";
(-e "$BWTINDEX.fa.fai")  or die "$BWTINDEX: bowtie index does not exist.\n";
(-d "$SYSCALL") or die "$SYSCALL: syscall directory does not exist.\n";

my $ANNOT;
my $SNPS;

while (@ARGV) {
  my $arg = shift;
  if ($arg=~/^\-annot/) {
    $ANNOT = shift;
  } elsif ($arg=~/^\-snp/) {
    $SNPS = shift;
  } elsif ($arg=~/\-verbose/) {
    $verbose = 1;
  } elsif ($arg=~/^\-dir/) {
    $work = shift;
  }
}

my $worktmp = "$work/tmp";
my $worklog = "$work/logs";

if (($work ne ".") && ($work ne "./")) { mkdir "$work"; }
mkdir "$worktmp";
mkdir "$worklog";
 
open(L, ">$work/run.log");

### Determine variable sites


my $SAMFILE = "$BAMFILE.sam";
print L "+samtools view $BAMFILE > $SAMFILE\n";
!system("samtools view $BAMFILE > $SAMFILE")  
        or die "samtools view command failed.\n"; 

print L "+samtools mpileup -A -BQ0 -d1000000 -f $GENOME $BAMFILE | " .
        "perl trim_mpileup.pl | " .
        "perl filter_vs.pl $R $P > $worktmp/prevar_sites 2> $worklog/filter_vs.log\n";
!system("samtools mpileup -A -BQ0 -d1000000 -f $GENOME $BAMFILE | " .
        "perl trim_mpileup.pl | " .
        "perl filter_vs.pl $R $P > $worktmp/prevar_sites 2> $worklog/filter_vs.log")
        or die "samtools pileup command failed.\n";
 
### Gene validation: generate (annotated) pileup output containing within-gene sites only  
### Generates annotate pileup (adds one colums at the end, with the ori sign)
if (defined($ANNOT) && (-e $ANNOT)) {
  print L "+intersect-genes $ANNOT $worktmp/prevar_sites -a " .
          "> $worktmp/var_sites 2> $worklog/intersect-genes.log\n" .
  !system("intersect-genes $ANNOT $worktmp/prevar_sites -a " .
          "> $worktmp/var_sites 2> $worklog/intersect-genes.log")
          or die "intersect-genes command failed.\n";
} else {
  print L "+perl add_sign.pl '.' < $worktmp/prevar_sites > $worktmp/var_sites\n"; 
  !system("perl add_sign.pl '.' < $worktmp/prevar_sites > $worktmp/var_sites")
          or die "could not annotate var_sites.\n";
}

### SNP validation: generate pileup files with extensions RDD and SNP
if (defined($SNPS) && (-e $SNPS)) {
  print L "+perl validate_snps.pl $worktmp/var_sites $SNPS -i dbSNP " .
          "-p $worktmp/var_sites 2> $worklog/validate_snps.log";
  !system("perl validate_snps.pl $worktmp/var_sites $SNPS -i dbSNP " .
          "-p $worktmp/var_sites 2> $worklog/validate_snps.log")
          or die "validate_snps command failed.\n";
} else {
  print L "+cp $worktmp/var_sites $worktmp/var_sites.RDD; open $worktmp/var_sites.SNP\n";
  !system("cp $worktmp/var_sites $worktmp/var_sites.RDD") or die "Could not create file $worktmp/var_sites.RDD\n"; 
  open(F, ">$worktmp/var_sites.SNP") or die "Could not create file $worktmp/var_sites.SNP\n";
  close(F);
}
   
 
### Apply sequencing error filters
###
### 1. sequencing single-strand filter
print L "+perl filter_strand_1.pl < $worktmp/var_sites " .
        "> $worktmp/var_sites.ds 2> $worklog/filter_strand_1.log\n";
!system("perl filter_strand_1.pl < $worktmp/var_sites " .
        "> $worktmp/var_sites.ds 2> $worklog/filter_strand_1.log")
        or die "filter_strand_1 command failed.\n";

### 2. read context filter
print L "+confirm-pos $worktmp/var_sites.ds $SAMFILE " .
        "> $worktmp/var_sites.ds.pos 2> $worklog/confirm_pos.log\n";
!system("confirm-pos $worktmp/var_sites.ds $SAMFILE " .
        "> $worktmp/var_sites.ds.pos 2> $worklog/confirm_pos.log")
        or die "confirm_pos command failed.\n";

### 3. variant and reference single-strand filter
print L "+perl filter_strand_2.pl < $worktmp/var_sites.ds.pos " .
        "> $worktmp/var_sites.ds.pos.vds 2> $worklog/filter_strand_2.log\n";
!system("perl filter_strand_2.pl < $worktmp/var_sites.ds.pos " .
        "> $worktmp/var_sites.ds.pos.vds 2> $worklog/filter_strand_2.log")
        or die "filter_strand_2 command failed.\n";

### 4. systematic error filter (Syscall)

print L "+perl make_pos_table.pl $GENOME < $worktmp/var_sites.ds.pos.vds " .
        "> $worktmp/var_sites.sites 2> $worklog/make_pos_table.log\n";
!system("perl make_pos_table.pl $GENOME < $worktmp/var_sites.ds.pos.vds " .
        "> $worktmp/var_sites.sites 2> $worklog/make_pos_table.log")
        or die "make_pos_table command failed.\n";
print L "+$SYSCALL/SysCall.pl $worktmp/var_sites.sites $SAMFILE " .
        "$worktmp/var_sites.syscall $SYSCALL/ > $worklog/syscall.log\n";
!system("perl $SYSCALL/SysCall.pl $worktmp/var_sites.sites $SAMFILE " .
        "$worktmp/var_sites.syscall $SYSCALL/ > $worklog/syscall.log")
        or die "SysCall command failed.\n";
print L "+perl filter_syscall.pl $worktmp/var_sites.syscall.heterozygous " .
        "< $worktmp/var_sites.ds.pos.vds > $worktmp/var_sites.ds.pos.vds.syscall\n";
!system("perl filter_syscall.pl $worktmp/var_sites.syscall.heterozygous " .
        "< $worktmp/var_sites.ds.pos.vds > $worktmp/var_sites.ds.pos.vds.syscall\n")
        or die "filter_syscall command failed.\n";


### SNP, annotation, and paralog filters
###
### 5. paralog filter
if (defined($ANNOT)) {
  print L "+extract-site-seq $worktmp/var_sites.ds.pos.vds.syscall $GENOME -annot $ANNOT " .
          "> $worktmp/var_sites.ds.pos.vds.syscall.seqs 2> $worklog/extract-site-seq.log\n";
  !system("extract-site-seq $worktmp/var_sites.ds.pos.vds.syscall $GENOME -annot $ANNOT " .
          "> $worktmp/var_sites.ds.pos.vds.syscall.seqs 2> $worklog/extract-site-seq.log")
          or die "extract-site-seq command (with annot) failed.\n";
} else { 
  # no annotation file given, extract genomic sequences
  print L "+extract-site-seq $worktmp/var_sites.ds.pos.vds.syscall $GENOME " .
          "> $worktmp/var_sites.ds.pos.vds.syscall.seqs 2> $worklog/extract-site-seq.log\n";
  !system("extract-site-seq $worktmp/var_sites.ds.pos.vds.syscall $GENOME " .
          "> $worktmp/var_sites.ds.pos.vds.syscall.seqs 2> $worklog/extract-site-seq.log")
          or die "extract-site-seq command (without annot) failed.\n";
}
print L "+bowtie -k 10 -f $BWTINDEX $worktmp/var_sites.ds.pos.vds.syscall.seqs " .
        "> $worktmp/var_sites.ds.pos.vds.syscall.seqs.bowtie 2> $worklog/bowtie.log\n"; 
!system("bowtie -k 10 -f $BWTINDEX $worktmp/var_sites.ds.pos.vds.syscall.seqs " .
        "> $worktmp/var_sites.ds.pos.vds.syscall.seqs.bowtie 2> $worklog/bowtie.log")
        or die "bowtie command failed.\n";
print L "+perl filter_paralog.pl $worktmp/var_sites.ds.pos.vds.syscall.seqs.bowtie $GENOME " .
        "< $worktmp/var_sites.ds.pos.vds.syscall " .
        "> $worktmp/var_sites.ds.pos.vds.syscall.nopar 2> $worklog/filter_paralog.log\n";
!system("perl filter_paralog.pl $worktmp/var_sites.ds.pos.vds.syscall.seqs.bowtie $GENOME " .
        "< $worktmp/var_sites.ds.pos.vds.syscall " .
        "> $worktmp/var_sites.ds.pos.vds.syscall.nopar 2> $worklog/filter_paralog.log")
        or die "filter_paralog command failed.\n";

### 6. annotation position filter
if (defined($ANNOT) && (-e $ANNOT)) {
  print L "+perl filter_near_junction.pl $ANNOT 3 " .
          "< $worktmp/var_sites.ds.pos.vds.syscall.nopar " .
          "> $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct\n";
  !system("perl filter_near_junction.pl $ANNOT 3 " .
          "< $worktmp/var_sites.ds.pos.vds.syscall.nopar " .
          "> $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct")
          or die "filter_near_junction command failed.\n";
} else {
  print L "No near-junction filter (no annotation).\n";
  !system("cp $worktmp/var_sites.ds.pos.vds.syscall.nopar $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct")
          or die "copy_near_junction command failed.\n";
}

### Separate sites into RDDs and SNPs
print L "+perl intersect.pl $worktmp/var_sites.RDD -v < $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct " . 
        ">$worktmp/var_sites.ds.pos.vds.syscall.nopar.njct.RDD 2> $worklog/intersect.RDD.log\n";
!system("perl intersect.pl $worktmp/var_sites.RDD -v < $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct " .
        "> $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct.RDD 2> $worklog/intersect.RDD.log")
        or die "intersect(RDD) command failed.\n";

print L "+perl intersect.pl $worktmp/var_sites.SNP -v < $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct." .
        ">$worktmp/var_sites.ds.pos.vds.syscall.nopar.njct.SNP 2> $worklog/intersect.SNP.log\n";
!system("perl intersect.pl $worktmp/var_sites.SNP -v < $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct " .
        ">$worktmp/var_sites.ds.pos.vds.syscall.nopar.njct.SNP 2> $worklog/intersect.SNP.log")
        or die "intersect(SNP) command failed\n";


### Convert final pileup to 'sites' format
print L "+perl apileup2sites.pl < $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct.RDD " . 
        "> $worktmp/sites.RDD.annot\n";
!system("perl apileup2sites.pl < $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct.RDD " .
        "> $worktmp/sites.RDD.annot") or die "apileup2sites(RDD) command failed.\n";
print L "+perl apileup2sites.pl < $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct.SNP " .
        "> $worktmp/sites.SNP.annot\n";
!system("perl apileup2sites.pl < $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct.SNP " .
        "> $worktmp/sites.SNP.annot") or die "apileup2sites(SNP) command failed.\n";


### Copy sites files into the work directory
!system("cp $worktmp/sites.RDD.annot $work/")
        or die "file copy operation failed.\n";
!system("cp $worktmp/sites.SNP.annot $work/")
        or die "SNP copy operation failed.\n";
!system("cut -f1-5 $work/sites.RDD.annot > $work/sites.RDD");
!system("cut -f1-5 $work/sites.SNP.annot > $work/sites.SNP");


### Collect stats on effects of filters
my ($log,$st);
$log  = "Sites                         RDD\tSNP\n";
$log .= "----------------------------  ---\t---\n";
$log .= "Variable sites:               ";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites | \
       perl intersect.pl $worktmp/var_sites.RDD`; chomp $st; $log .= "\t$st";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites | \
       perl intersect.pl $worktmp/var_sites.SNP`; chomp $st; $log .= "\t$st\n";

$log .= "Single-strand filter:         ";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds | \
       perl intersect.pl $worktmp/var_sites.RDD`; chomp $st; $log .= "\t$st";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds | \
       perl intersect.pl $worktmp/var_sites.SNP`; chomp $st; $log .= "\t$st\n";

$log .= "Read-context filter:          ";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds.pos | \
       perl intersect.pl $worktmp/var_sites.RDD`; chomp $st; $log .= "\t$st";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds.pos | \
       perl intersect.pl $worktmp/var_sites.SNP`; chomp $st; $log .= "\t$st\n";

$log .= "Ref-Var strand filter:        ";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds.pos.vds | \
       perl intersect.pl $worktmp/var_sites.RDD`; chomp $st; $log .= "\t$st";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds.pos.vds | \
       perl intersect.pl $worktmp/var_sites.SNP`; chomp $st; $log .= "\t$st\n";

$log .= "SysCall filter:               ";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds.pos.vds.syscall | \
       perl intersect.pl $worktmp/var_sites.RDD`; chomp $st; $log .= "\t$st";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds.pos.vds.syscall | \
       perl intersect.pl $worktmp/var_sites.SNP`; chomp $st; $log .= "\t$st\n";

$log .= "Paralog filter:               ";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds.pos.vds.syscall.nopar | \
       perl intersect.pl $worktmp/var_sites.RDD`; chomp $st; $log .= "\t$st";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds.pos.vds.syscall.nopar | \
       perl intersect.pl $worktmp/var_sites.SNP`; chomp $st; $log .= "\t$st\n";

$log .= "Near-junction filter (Final): ";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct | \
       perl intersect.pl $worktmp/var_sites.RDD`; chomp $st; $log .= "\t$st";
$st = `awk \'{ if (\$7!=\"*\") print $_; }\' $worktmp/var_sites.ds.pos.vds.syscall.nopar.njct | \
       perl intersect.pl $worktmp/var_sites.SNP`; chomp $st; $log .= "\t$st\n";


if ($log ne "") {
  if (open(F, ">$work/filters.stats")) {
    print F $log;
    close(F);
  } else {
    print STDERR "could not open file filters.stats for writing.\n";
  }
}

close(L);

### Cleanup temporary and log files if not requested otherwise
if  (!$verbose) {
  system("rm -rf $worktmp");
  system("rm -rf $worklog");
}


exit(0);
