#include "libc.h"
#include "lib.h"

#define DEFAULT_ALLOC 10000

#define MAX_INTV_LEN  20000

static char Usage[] = "%s annotation mutations [-a]\n";

#ifdef MOUSE
char *Chrom[] = {"chr1", "chr2", "chr3", "chr4","chr5",
                 "chr6", "chr7", "chr8", "chr9", "chr10",
                 "chr11", "chr12", "chr13", "chr14","chr15",
                 "chr16", "chr17", "chr18", "chr19", "chrX",
                 "chrY", "chrM"};
int numChroms = 22;
#else
char *Chrom[] = {"chr1", "chr2", "chr3", "chr4","chr5",
                 "chr6", "chr7", "chr8", "chr9", "chr10",
                 "chr11", "chr12", "chr13", "chr14","chr15",
                 "chr16", "chr17", "chr18", "chr19", "chr20",
                 "chr21", "chr22", "chrX", "chrY", "chrM"};
int numChroms = 25;
#endif 

typedef struct {
   int  from, to, gaid, orig;
   char ori;
} interval_t;

typedef struct {
   int  pos, gaid;
   char ref, mut;
} mutation_t;


static char   errmsg[100];

static int    maxIntvLen = MAX_INTV_LEN;

static void loadIntervals(const char *filename, interval_t **ints, int *numints, int *maxints);
static int  lookup_chrom(char *subj, char *chroms[], int numchroms);
static int  lookup_ints(int pos, int idx, interval_t *ints, int start, int numints);
static int  intv_cmp(const void *a, const void *b);


static int  intv_cmp(const void *a, const void *b) {
   interval_t *A = (interval_t *)a;
   interval_t *B = (interval_t *)b;
   int t;

   if (t=A->gaid-B->gaid)
     return t;
   else if (t=A->from-B->from) 
     return t;
   else
     return (A->to-B->to);
}

void main(int argc, char *argv[])
{
    interval_t **Ints;
    int         *intsCounts;
    int          numInts, maxInts;
    mutation_t   mut;
    char         buf[100000], dummy[100], match_list[100000], *s;
    int          idxc, idxi, i;
    int          findAll = 0;
    FILE        *fp;
  
    if (argc<3) fatalf(Usage, argv[0]);

    if (argc==4) {
      if (!strcmp(argv[3], "-a"))
        findAll = 1;
      else 
        fatalf(Usage, argv[0]);
    }

    numInts = maxInts = 0;

    Ints = (interval_t **)malloc(25*sizeof(interval_t));
    memset(Ints, 0, 25*sizeof(interval_t *));

    loadIntervals(argv[1], &(Ints[0]), &numInts, &maxInts);

    qsort(Ints[0], numInts, sizeof(interval_t), intv_cmp); 

    intsCounts = (int *)malloc(25*sizeof(int));
    memset(intsCounts, 0, 25*sizeof(int));

    for (i=0; i<numInts; i++) {
      int gaid = Ints[0][i].gaid;
      if (Ints[gaid]==NULL) Ints[gaid] = Ints[0]+i; 
      intsCounts[gaid]++;
    }

    fp = ckopen(argv[2], "r");
    while (fgets(buf, 100000, fp)!=NULL) {
       int idummy;
       /* chr1    1250898 A       34      ,,,c,,,,..,,,,c,cccc,cccc,cccccc,c      HH>#FHFHHHDH5FHCHF#HHFHGHIFHHHDHGD */
       if (sscanf(buf, "%s\t%d\t%c\t%d\t%s\t",
                  dummy,&(mut.pos),&(mut.ref),&idummy,match_list)!=5) {
         fprintf(stderr, "Unrecognized mutation line. Skipping site. %s\n", buf);
         continue;
       }

       if ((idxc=lookup_chrom(dummy, Chrom, numChroms))<0) {
         fprintf(stderr, "Unrecognized chromosome number. Skipping site. %s\n", buf);
         continue;
       }

       for (s=match_list; *s; s++) {
         char ch = toupper(*s);
         if (ch=='A' || ch=='C' || ch=='G' || ch=='T') {
            mut.mut = ch; break;
         }
       }

       if (!findAll) {
         if ((idxi=lookup_ints(mut.pos, idxc, Ints[idxc], 0, intsCounts[idxc]))>=0) {
           int l = strlen(buf);
           if (buf[l-1] == '\n') buf[l-1] = '\0';
           printf("%s\t%c\n", buf, (Ints[idxc]+idxi)->ori);
         }
       } else {
         int  next = 0;
         char dbgstr[1000], dbgori;

         dbgstr[0] = '\0'; dbgori = '.';

         while ((idxi=lookup_ints(mut.pos, idxc, Ints[idxc], next, intsCounts[idxc]))>=0) {

           sprintf(dbgstr+strlen(dbgstr), ",%d", (Ints[idxc]+idxi)->orig); 

           if (((dbgori == '+') && ((Ints[idxc]+idxi)->ori=='-')) ||
               ((dbgori == '-') && ((Ints[idxc]+idxi)->ori=='+')))
             dbgori = '*';
           else if (dbgori == '*')
             ;
           else 
             dbgori = (Ints[idxc]+idxi)->ori;

           next = idxi+1; 
         }
         if (*dbgstr) {
           int l = strlen(buf);
           if (buf[l-1] == '\n') buf[l-1] = '\0';
           printf("%s\t%c\n", buf, dbgori);
         }
       }
    }
    fclose(fp);

    free(Ints);

    exit(0);
}

static void loadIntervals(const char *filename, interval_t **ints, int *numints, int *maxints)
{
   FILE *fp = ckopen(filename, "r");
   char  buf[100000], dummy[100], dummy1[100], chrom[100], ori;
   int   from, to, i, count; 

   *numints = *maxints = 0;

   count = 0;

   while (fgets(buf, 100000, fp)!=NULL) {
      if (strstr(buf, "\texon\t")==NULL) {
        count++; continue;
      }

      if (sscanf(buf, "%s\t%s\texon\t%d\t%d\t%s\t%c\t", chrom, dummy, &from, &to, dummy1, &ori)!=6) {
        fprintf(stderr, "Unrecognized exon line. Skipping. %s\n", buf);
        count++;
        continue;
      }
   
      if ((ori!='+') && (ori!='-')) {
        fprintf(stderr, "no strand information. Skipping exon. %s\n", buf);
        count++;
        continue;
      }

      if ((i=lookup_chrom(chrom, Chrom, numChroms))<0) {
        fprintf(stderr, "cannot find chromosome. Skipping interval. %s\n", buf); 
        count++;
        continue;
      }

      if (*numints>=*maxints) {
        *maxints = max(DEFAULT_ALLOC, 2*(*maxints));
        *ints = realloc(*ints, (*maxints)*sizeof(interval_t));
      }
      (*ints)[*numints].from = from;
      (*ints)[*numints].to = to; 
      (*ints)[*numints].gaid = i;
      (*ints)[*numints].ori = ori;
      (*ints)[*numints].orig = count;

      (*numints)++;
     
      if (to-from+1 > maxIntvLen) maxIntvLen = to-from+1;

      count++;
   }
   
   fclose(fp);
}

static int lookup_chrom(char *subj, char *chroms[], int numchroms)
{
   int i;

   for (i=0; (i<numchroms) && strcmp(chroms[i],subj); i++);
   return (i==numchroms) ? -1 : i;
}

static int lookup_ints(int pos, int idx, interval_t *ints, int start, int numints)
{
   int i;
 
   for (i=start; i<numints; i++) {
     if (ints[i].gaid!=idx) {
       sprintf(errmsg, "Incorrect gaid in lookup list. Found %d, expecting %d.\n", ints[i].gaid, idx);
       fatalf("%s", errmsg);
     }

     if ((ints[i].from<=pos) && (pos<=ints[i].to)) return i;

     if (pos<ints[i].from) return -1;
   }
     
   return -1;
}
